import os
import json
from PIL import Image, ImageStat
import random
import argparse
from pathlib import Path
import csv

random.seed(1337)

def flatten(t):
    return [item for sublist in t for item in sublist]

def hash_image(image_path):
    img = Image.open(image_path).resize((8, 8), Image.LANCZOS).convert(mode="L")
    mean = ImageStat.Stat(img).mean[0]
    return sum((1 if p > mean else 0) << i for i, p in enumerate(img.getdata()))


def build_ignore_list(out_path_dpl, out_path_il, corpus_path):

    ignore_list = []
    duplicate_dict = {}
    counter = 0

    if os.path.exists(out_path_dpl):
        duplicate_dict = json.load(open(os.path.join(out_path_dpl)))

    else:
        image_folder_path = Path(corpus_path) / 'Manually_Annotated_Images'
        exclude_extensions = ["db"]
        for root, dir, filenames in os.walk(image_folder_path):
            for f in filenames:
                if f.split(".")[-1] in exclude_extensions:
                    continue
                if counter % 1000 == 0:
                    print("hashing image {}".format(counter))
                counter += 1
                path = os.path.join(root, f)
                hash = hash_image(path)
                if hash not in duplicate_dict.keys():
                    duplicate_dict[hash] = [root.split(os.sep)[-1] + "/" + f]
                else:
                    duplicate_dict[hash].append(root.split(os.sep)[-1] + "/" + f)

        with open(out_path_dpl, "w") as fp:
            json.dump(duplicate_dict, fp, sort_keys=True, indent=4)
            exit()

    dupl = [x for x in duplicate_dict.values() if len(x) > 1]

    # Do not touch validation samples
    print('n duplicates: {}'.format(len(flatten(dupl))))
    validation_samples = []
    print('Removing validation samples')
    with open(Path(corpus_path) / 'Manually_Annotated_file_lists' / 'validation.csv', 'r') as validation_file:
        labels = csv.reader(validation_file, delimiter=',')
        validation_samples = [x[0] for x in list(labels)]

    dupl = [list(filter(lambda x: x not in validation_samples, d)) for d in dupl]
    print('n duplicates: {}'.format(len(flatten(dupl))))


    # Keeping one random image
    for img_list in dupl:
        random.shuffle(img_list)
        for i in range(len(img_list) - 1):
            ignore_list.append(img_list[i])
    with open(out_path_il, "w") as fp:
        json.dump(ignore_list, fp, sort_keys=True, indent=4)


if __name__ == "__main__":
    # Parse Arguments
    my_parser = argparse.ArgumentParser()
    my_parser.add_argument("--out_path_dpl", help="Path to store the caluclated hashes")
    my_parser.add_argument("--out_path_il", help="Path to store the final ignore list")
    my_parser.add_argument(
        "--coprpus_dir", help="Path to the affect net corpus directory"
    )
    args = my_parser.parse_args()

    build_ignore_list(args.out_path_dpl, args.out_path_il, args.coprpus_dir)



import pandas as pd
from abc import ABC, abstractmethod


class Statistics(ABC):
    @property
    @abstractmethod
    def info(self):
        ...

    def _populate_stats(self, data: pd.DataFrame, split="train"):
        """
        Adding statistic metadata to the dataset information. Needs to manually called in the dataset child class.
        Args:
            data: The input data to analyze. Could be of type: pandas.DataFrame, ...
            split: The splitname that is set in the dictionary.
        """
        if isinstance(data, pd.DataFrame):
            stats = self.stats_from_dataframe(data)
        else:
            raise TypeError()
        self.info.metadata[split] = stats

    @staticmethod
    def stats_from_dataframe(df: pd.DataFrame) -> dict:
        """
        Calculates statistics for the given dataframe.
        The exact form of the return value depens on the input. The function will generate individual stats for each column in the dataframe.
        If the column is of dtype objects the calculated attributes are: count, uniqe, top, freq, dist.
        Else the statistic comprises the attributes: count, mean, std, min, max, percentiles 25%, 50%,75%.
        See pandas.dataframe.describe() for more information

        Esample output:
        {
        'expression':
            {'count': 382670,
            'unique': 11,
            'top': 1,
            'freq': 125245,
            'dist': {'1': 125245, '10': 75505, '0': 69841, '8': 30753, '2': 22690, '6': 22381, '3': 13077, '9': 10722, '4': 5581, '7': 3467, '5': 3408}
            },
        'arousal': {'count': 296443.0, 'mean': 0.08697717502906124, 'std': 0.32627918341864814, 'min': -0.998677, '25%': -0.085034, '50%': 0.0555556, '75%': 0.216057, 'max': 0.996903},
        'valence': {'count': 296443.0, 'mean': 0.1872163682477575, 'std': 0.5039924392122456, 'min': -0.999081, '25%': -0.161064, '50%': 0.203252, '75%': 0.626984, 'max': 0.999081}
        }

        Args:
            df: Input dataframe

        Returns:
            dict: Dictionary containing the calculated statistics

        """
        ret = {}

        # Categorical stats
        df_cat = df.select_dtypes(include=["object"])
        if not df_cat.empty:
            dsc_cat = df_cat.describe().to_dict()
            distribution = {
                l: {"dist": df_cat[l].value_counts(sort=True).to_dict()}
                for l in df_cat.head()
            }
            for k, v in distribution.items():
                dsc_cat[k].update(v)
            ret.update(dsc_cat)

        # Continuous stats
        df_cont = df.select_dtypes(exclude=["object"])
        if not df_cont.empty:
            dsc_cont = df_cont.describe().to_dict()
            ret.update(dsc_cont)

        return ret



from enum import Enum
import numpy as np

class DataTypes(Enum):
  VIDEO = 0
  AUDIO  = 1
  FEATURE = 2


class AnnoTypes(Enum):
  DISCRETE = 0
  CONTINOUS = 1
  FREE = 2
  POINT = 3
  DISCRETE_POLYGON = 4


'''Helper'''
def string_to_enum(enum, string):
  for e in enum:
    if e.name == string.upper():
      return e
  raise ValueError('{} not part of enumeration  {}'.format(string, enum))



import xml.etree.ElementTree as ET
from hcai_datasets.hcai_nova_dynamic.utils.ssi_data_types import FileTypes, string_to_enum
from enum import Enum
import numpy as np
import csv
from struct import *


class SchemeTypes(Enum):
    DISCRETE = 0
    CONTINUOUS = 1
    FREE = 2

class Scheme:
    #def __init__(self, name="", type="", sr=0, min=0, max=0):
    def __init__(self):
        self.name = ""
        self.type = ""
        self.sr = 0
        self.min = 0
        self.max = 0
        self.classes = {}

class Anno:
    #def __init__(self, ftype="UNDEF", size=0, role="", annotator=""):
    def __init__(self):    
        self.ftype = string_to_enum(FileTypes, "UNDEF")
        self.size = 0
        self.role = ""
        self.annotator = ""
        self.data = None
        self.scheme = Scheme()

    def load_header(self, path):
        tree = ET.parse(path)
        root = tree.getroot()

        for child in root:
            for key,val in child.attrib.items():
                if child.tag == 'info':  
                    if key == 'ftype':
                        self.ftype = string_to_enum(FileTypes, val)
                    elif key == 'size':
                        self.size = int(val)       
                elif child.tag == 'meta':
                    if key == 'role':
                        self.role = val
                    elif key == 'annotator':
                        self.annotator = val
                elif child.tag == 'scheme':
                    if key == 'name':
                        self.scheme.name = val
                    if key == 'type':
                        self.scheme.type = string_to_enum(SchemeTypes, val)
                    if key == 'sr':
                        self.scheme.sr = float(val)
                    if key == 'min':
                        self.scheme.min = int(val)
                    if key == 'max':
                        self.scheme.max = int(val)
            if child.tag == 'scheme' and self.scheme.type == SchemeTypes.DISCRETE :
                for item in child:              
                    for key,val in item.attrib.items():
                        if key == "name":
                            class_name = val
                        elif key == "id":
                            id = val
                    self.scheme.classes[id] = class_name

    def load_data_discrete(self, path):
        dt = {'names':('from', 'to', 'class', 'confidence'),
                          'formats':('f8', 'f8', 'i4', 'f4')}
        if self.ftype == FileTypes.ASCII:
            self.data = np.loadtxt(path, dtype=dt, delimiter=';')
        elif self.ftype == FileTypes.BINARY:
            self.data = np.fromfile(path, dtype=dt)
        else:
            raise ValueError('FileType {} not supported'.format(self.ftype))

    def load_data_continuous(self, path):
        dt = {'names':('score', 'confidence'),
                          'formats':('f4', 'f4')}
        if self.ftype == FileTypes.ASCII:
            self.data = np.loadtxt(path, dtype=dt, delimiter=';')
        elif self.ftype == FileTypes.BINARY:
            self.data = np.fromfile(path, dtype=dt)
        else:
            raise ValueError('FileType {} not supported'.format(self.ftype))

    def load_data_free(self, path):

        data = []

        if self.ftype == FileTypes.ASCII:
            with open(path, "r") as ascii_file:
                ascii_file_reader = csv.reader(ascii_file, delimiter=';', quotechar='"')
                for row in ascii_file_reader:
                    f = float(row[0])
                    t = float(row[1])
                    l = row[2]
                    c = float(row[3])
                    if not l in self.scheme.classes.values():
                        if len(self.scheme.classes) == 0:
                            v = 0
                        else:
                            v = max(self.scheme.classes.keys()) + 1 
                        self.scheme.classes[v] = l
                    data.append((f,t,v,c))

        elif self.ftype == FileTypes.BINARY:
            with open(path, "rb") as binary_file:
                counter = 0
                binary_file.seek(0)  
                
                while counter < self.size:
                    #from (8byte float)
                    f = unpack('d', binary_file.read(8))[0]
                    ##to (8byte float)
                    t = unpack('d', binary_file.read(8))[0]
                    #length of label (4byte uint)
                    lol = unpack('i', binary_file.read(4))[0]
                    #the label (lol * byte)
                    l = binary_file.read(lol).decode("ISO-8859-1") 
                    #confidence (4Byte float)
                    c = unpack('f', binary_file.read(4))[0]
                 
                    if not l in self.scheme.classes.values():
                        if len(self.scheme.classes) == 0:
                            v = 0
                        else:
                            v = max(self.scheme.classes.keys()) + 1 
                        self.scheme.classes[v] = l
                    data.append((f,t,v,c))
                    counter += 1
        else:
            raise ValueError('FileType {} not supported'.format(self.ftype))

        dt = {'names':('from', 'to', 'label_id', 'confidence'),
                          'formats':('f8', 'f8', 'i4', 'f4')}
        self.data = np.array(data, dt)

    def load_data(self, path):
        if self.scheme.type == SchemeTypes.DISCRETE:
            self.load_data_discrete(path)
        elif self.scheme.type == SchemeTypes.CONTINUOUS:
            self.load_data_continuous(path)
        elif self.scheme.type == SchemeTypes.FREE:
            self.load_data_free(path)
        else:
            raise ValueError('SchemeType {} not supported'.format(self.scheme.type))
        return self.data

    def load(self, path):
        self.load_header(path)
        self.load_data(path + '~')
    
if __name__ == "__main__":

    '''discrete annotations'''
    anno_discrete_ascii = Anno()
    anno_discrete_ascii.load("Testfiles/discrete_ascii.annotation")

    anno_discrete_binary = Anno()
    anno_discrete_binary.load("Testfiles/discrete_binary.annotation")

    '''continous annotations'''
    anno_cont_ascii = Anno()
    anno_cont_ascii.load("Testfiles/continuous_ascii.annotation")

    anno_cont_binary = Anno()
    anno_cont_binary.load("Testfiles/continuous_binary.annotation")
    
    '''free annotations'''
    anno_free_ascii = Anno()
    anno_free_ascii.load("Testfiles/free_ascii.annotation")

    anno_free_binary = Anno()
    anno_free_binary.load("Testfiles/free_binary.annotation")
  
    ...



"""hcai_ckplus dataset."""
import tensorflow as tf
import tensorflow_datasets as tfds
import pandas as pd
from tensorflow_datasets.core.splits import Split
from pathlib import Path
from hcai_dataset_utils.statistics import Statistics


_DESCRIPTION = """
The Extended Cohn-Kanade (CK+) dataset contains 593 video sequences from a total of 123 different subjects, ranging from 18 to 50 years of age with a variety of genders and heritage.
"""

_CITATION = """
@inproceedings{lucey2010extended,
  title={The extended cohn-kanade dataset (ck+): A complete dataset for action unit and emotion-specified expression},
  author={Lucey, Patrick and Cohn, Jeffrey F and Kanade, Takeo and Saragih, Jason and Ambadar, Zara and Matthews, Iain},
  booktitle={2010 ieee computer society conference on computer vision and pattern recognition-workshops},
  pages={94--101},
  year={2010},
  organization={IEEE}
}
"""


class HcaiCkplus(tfds.core.GeneratorBasedBuilder, Statistics):
    """DatasetBuilder for hcai_ckplus dataset."""

    VERSION = tfds.core.Version("1.0.0")
    RELEASE_NOTES = {
        "1.0.0": "Initial release.",
    }

    LABELS = [
        "neutral",
        "anger",
        "contempt",
        "disgust",
        "fear",
        "happy",
        "sadness",
        "suprise",
    ]

    def __init__(self, *, dataset_dir=None, **kwargs):
        self.dataset_dir = Path(dataset_dir)
        super(HcaiCkplus, self).__init__(**kwargs)

    def _info(self) -> tfds.core.DatasetInfo:
        """Returns the dataset metadata."""
        return tfds.core.DatasetInfo(
            builder=self,
            description=_DESCRIPTION,
            metadata=tfds.core.MetadataDict({}),
            features=tfds.features.FeaturesDict(
                {
                    "image": tfds.features.Image(shape=(None, None, 3)),
                    # 0=neutral, 1=anger, 2=contempt, 3=disgust, 4=fear, 5=happy, 6=sadness, 7=surprise)
                    "label": tfds.features.ClassLabel(names=self.LABELS),
                    "rel_file_path": tf.string,
                }
            ),
            # If there's a common (input, target) tuple from the
            # features, specify them here. They'll be used if
            # `as_supervised=True` in `builder.as_dataset`.
            supervised_keys=("image", "label"),  # Set to `None` to disable
            homepage="https://dataset-homepage/",
            citation=_CITATION,
        )

    def _populate_meta_data(self, data):
        df = pd.DataFrame(data, columns=["file_name", "emotion"])
        df = df.drop(columns="file_name")
        self._populate_stats(df)

    def _split_generators(self, dl_manager: tfds.download.DownloadManager):
        """Returns SplitGenerators."""

        emo_anno_files = list(self.dataset_dir.glob("Emotion/**/**/*.txt"))
        samples = []

        for ef in emo_anno_files:
            with open(ef, "r") as f:
                emotion = self.LABELS[int(float(f.readline().strip()))]
                fn_img = ef.stem.replace("_emotion", ".png")
                samples.append((fn_img, emotion))

                # add neutral images
                fn_img_neut = ef.stem.rsplit("_", 2)[0] + "_00000001.png"
                samples.append((fn_img_neut, "neutral"))

        self._populate_meta_data(samples)
        return {Split.TRAIN: self._generate_examples(samples)}

    def _generate_examples(self, files):
        """Yields examples."""

        for f, e in files:
            rel_path = (Path(*f.split("_")[:-1]) / f).resolve()
            yield str(f), {
                "image": self.dataset_dir / "cohn-kanade-images" / rel_path,
                "label": e,
                "rel_file_path": str(rel_path),
            }



import cv2
import numpy as np
import os
import tensorflow_datasets as tfds
import tensorflow as tf
from hcai_datasets.hcai_nova_dynamic.utils import nova_types as nt
from hcai_datasets.hcai_nova_dynamic.utils.ssi_stream_utils import Stream
from hcai_datasets.hcai_nova_dynamic.utils.nova_utils import merge_role_key, split_role_key
from typing import Union
from abc import ABC, abstractmethod

class Data(ABC):

    lazy_connector = tfds.features.FeaturesDict({
        'frame_start': tf.dtypes.float32,
        'frame_end': tf.dtypes.float32,
        'file_path': tfds.features.Text()
    })


    def __init__(self,  role: str = '', name: str = '', file_ext: str = 'stream', sr: int = 0, data_type: nt.DataTypes = None, is_valid: bool = True, sample_data_path: str = '', lazy_loading: bool = False):
        self.role = role
        self.name = name
        self.is_valid = is_valid
        self.sr = sr
        self.file_ext = file_ext
        self.lazy_loading = lazy_loading
        self.data_type = data_type

        # Set when populate_meta_info is called
        self.sample_data_shape = None
        self.tf_data_type = None
        self.meta_loaded = False

        # Set when open_file_reader is called
        self.file_path = None
        self.file_reader = None
        self.dur = None

        if sample_data_path:
            self.populate_meta_info(sample_data_path)

    def data_stream_opend(self):
        if not self.file_reader:
            print('No datastream opened for {}'.format(merge_role_key(self.role, self.name)))
            raise RuntimeError('Datastream not loaded')


    def get_tf_info(self):
        if self.meta_loaded:
            if self.lazy_loading:
                feature_connector = self.lazy_connector
                key = merge_role_key(self.role, self.name)
                return (key, feature_connector)
            else:
                return self.get_tf_info_hook()
        else:
            print('Meta data has not been loaded for file {}. Call get_meta_info() first.'.format(
                merge_role_key(self.role, self.name)))


    def get_sample(self, frame_start: int, frame_end: int):
        """
        Returns the sample for the respective frames. If lazy loading is true, only the filepath and frame_start, frame_end will be returned.
        """
        if self.lazy_loading:
            return {
                'frame_start': frame_start,
                'frame_end': frame_end,
                'file_path': self.file_path
            }
        else:
            return self.get_sample_hook(frame_start, frame_end)

    def open_file_reader(self, path: str):
        self.file_path = path
        self.open_file_reader_hook(path)

    @abstractmethod
    def get_tf_info_hook(self):
        """
        Returns the features for this datastream to create the DatasetInfo for tensorflow
        """
        ...

    @abstractmethod
    def get_sample_hook(self, start_frame: int, end_frame: int):
        """
        Returns a data chunk from start frame to end frame
        """
        ...

    @abstractmethod
    def open_file_reader_hook(self, path: str):
        """
        Opens a filereader for the respective datastream. Sets attributes self.file_reader and self.dur
        """
        ...

    @abstractmethod
    def populate_meta_info(self, path: str):
        """
        Opens a data sample from the provided path to extract additional data that is not in the database
        """
        ...

    @abstractmethod
    def close_file_reader(self):
        """
        Closes a filereader for the respective datastream
        """
        ...


class AudioData(Data):
    pass


class VideoData(Data):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_tf_info_hook(self) -> (str, tfds.features.Sequence):
        feature_connector = tfds.features.Sequence(
        tfds.features.Image(shape=self.sample_data_shape, dtype=np.uint8))
        return merge_role_key(self.role, self.name), feature_connector

    def get_sample_hook(self, frame_start_ms: int, frame_end_ms: int):
        start_frame = frame_start_ms / 1000 * self.sr
        end_frame = frame_end_ms / 1000 * self.sr
        length = int(end_frame - start_frame)

        self.file_reader.set(cv2.CAP_PROP_POS_FRAMES, start_frame)
        chunk = np.zeros((length, ) + self.sample_data_shape, dtype=np.uint8)

        for i in range(length):
            ret, frame = self.file_reader.read()

            if not ret:
                raise IndexError('Video frame {} out of range'.format(i))

            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
            chunk[i] = frame

        return chunk

    def open_file_reader_hook(self, path: str):
        if not os.path.isfile(path):
            print('Session file not found at {}.'.format(path))
            raise FileNotFoundError()
        self.file_reader = cv2.VideoCapture(path)
        fps = self.file_reader.get(cv2.CAP_PROP_FPS)  # OpenCV2 version 2 used "CV_CAP_PROP_FPS"
        frame_count = int(self.file_reader.get(cv2.CAP_PROP_FRAME_COUNT))
        self.dur = frame_count / fps

    def populate_meta_info(self, path: str):
        if not os.path.isfile(path):
            print('Sample file not found at {}. Can\'t load metadata.'.format(path))
            raise FileNotFoundError()
        file_reader = cv2.VideoCapture(path)
        width = int(file_reader.get(cv2.CAP_PROP_FRAME_WIDTH))
        height = int(file_reader.get(cv2.CAP_PROP_FRAME_HEIGHT))
        depth = 3
        self.sample_data_shape = (height, width, depth)
        self.meta_loaded = True

    def close_file_reader(self):
        return self.file_reader.release()


class StreamData(Data):
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_tf_info_hook(self) -> (str, tfds.features.Sequence):
        feature_connector = tfds.features.Sequence(tfds.features.Tensor(shape=self.sample_data_shape, dtype=self.tf_data_type))
        return merge_role_key(self.role, self.name), feature_connector

    def get_sample_hook(self, frame_start_ms: int, frame_end_ms: int):
        try:
            self.data_stream_opend()
            start_frame = milli_seconds_to_frame(self.sr, frame_start_ms)
            end_frame = milli_seconds_to_frame(self.sr, frame_end_ms)
            return self.file_reader.data[start_frame:end_frame]
        except RuntimeError:
            print('Could not get chunk {}-{} from data stream {}'.format(frame_start_ms, frame_end_ms, merge_role_key(self.role, self.name)))

    def open_file_reader_hook(self, path: str) -> bool:
        stream = Stream(path)
        if stream:
            self.file_reader = stream
            self.dur = stream.data.shape[0] / stream.sr
            return True
        else:
            print('Could not open Stream {}'.format(str))
            return False

    def close_file_reader(self):
        return True

    def populate_meta_info(self, path: str):
        stream = Stream().load_header(path)
        self.sample_data_shape = (stream.dim,)
        self.tf_data_type = stream.tftype
        self.meta_loaded = True

##########################
# General helper functions
##########################

def frame_to_seconds(sr: int, frame: int) -> float:
    return frame / sr


def seconds_to_frame(sr: int, time_s: float) -> int:
    return round(time_s * sr)


def milli_seconds_to_frame(sr: int, time_ms: int) -> int:
    return seconds_to_frame(sr=sr, time_s= time_ms / 1000)


def parse_time_string_to_ms(frame: Union[str, int, float]) -> int:
    # if frame is specified milliseconds as string
    if str(frame).endswith('ms'):
        try:
            return int(frame[:-2])
        except ValueError:
            raise ValueError('Invalid input format for frame in milliseconds: {}'.format(frame))
    # if frame is specified in seconds as string
    elif str(frame).endswith('s'):
        try:
            frame_s = float(frame[:-1])
            return int(frame_s * 1000)
        except ValueError:
            raise ValueError('Invalid input format for frame in seconds: {}'.format(frame))
    # if type is float we assume the input will be seconds
    elif isinstance(frame, float) or '.' in str(frame):
        try:
            print('WARNING: Automatically inferred type for frame {} is float.'.format(frame))
            return int(1000 * float(frame))
        except ValueError:
            raise ValueError('Invalid input format for frame: {}'.format(frame))
    # if type is int we assume the input will be milliseconds
    elif isinstance(frame, int):
        try:
            print('WARNING: Automatically inferred type for frame {} is int.'.format(frame))
            return int(frame)
        except ValueError:
            raise ValueError('Invalid input format for frame: {}'.format(frame))



"""hcai_nova_dynamic dataset."""

import numpy as np
import os
import shutil
import sys
import tensorflow_datasets as tfds
import tensorflow as tf

import hcai_datasets.hcai_nova_dynamic.utils.nova_types as nt
import hcai_datasets.hcai_nova_dynamic.utils.nova_data_utils as ndu
import hcai_datasets.hcai_nova_dynamic.utils.nova_anno_utils as nau

from hcai_datasets.hcai_nova_dynamic.utils.nova_data_utils import AudioData, VideoData, StreamData
from hcai_datasets.hcai_nova_dynamic.utils.nova_anno_utils import DiscreteAnnotation, ContinousAnnotation, FreeAnnotation

from hcai_datasets.hcai_nova_dynamic.nova_db_handler import NovaDBHandler
from hcai_datasets.hcai_nova_dynamic.utils.nova_utils import *

# TODO(hcai_audioset): Markdown description  that will appear on the catalog page.
_DESCRIPTION = """
The Nova Dynamic dataset can be used to retrieve the data and labels for a certain session or a certain part of a session of a nova dataset. 
This is part of the Nova CML Python backend (https://github.com/hcmlab/nova)
To specify which data to load use the following format: 

TODO: x
 
"""

# TODO(hcai_audioset): BibTeX citation
_CITATION = """
"""


class HcaiNovaDynamic(tfds.core.GeneratorBasedBuilder):
    """DatasetBuilder for hcai_nova_dynamic dataset."""

    VERSION = tfds.core.Version('1.0.0')
    RELEASE_NOTES = {
        '1.0.0': 'Initial release.',
    }

    def __init__(self, *, db_config_path=None, db_config_dict=None, dataset=None, nova_data_dir=None, sessions=None,
                 annotator=None,
                 schemes=None, roles=None, data_streams=None, start=None, end=None, left_context='0', right_context='0',
                 frame_size='1', stride=None,
                 flatten_samples=False, supervised_keys=None, clear_cache=True, add_rest_class=True, lazy_loading=False,
                 **kwargs):
        """
        Initialize the HcaiNovaDynamic dataset builder
        Args:
          nova_data_dir: the directory to look for data. same as the directory specified in the nova gui.
          frame_size: the framesize to look at. the matching annotation will be calculated as majority vote from all annotations that are overlapping with the timeframe.
          left_context: additional data to pass to the classifier on the left side of the frame.
          right_context: additional data to pass to the classifier on the left side of the frame.
          stride: how much a frame is moved to calculate the next sample. equals framesize by default.
          flatten_samples: if set to True samples with the same annotation scheme but from different roles will be treated as separate samples. only <scheme> is used for the keys.
          supervised_keys: if specified the dataset can be used with "as_supervised" set to True. Should be in the format <role>.<scheme>. if flatten_samples is true <role> will be ignored.
          clear_cache:  when set to True the cache will be cleared else the cached dataset will be used. make sure that dataset and sample config did not change. defaults to true.
          add_rest_class: when set to True an additional restclass will be added to the end the label list
          db_config_path: path to a configfile whith the nova database config.
          db_config_dict: dictionary with the nova database config. can be used instead of db_config_path. if both are specified db_config_dict is used.
          dataset: the name of the dataset. must match the dataset name in the nova database.
          sessions: list of sessions that should be loaded. must match the session names in nova.
          annotator: the name of the annotator that labeld the session. must match annotator names in nova.
          schemes: list of the annotation schemes to fetch
          roles: list of roles for which the annotation should be loaded.
          data_streams: list datastreams for which the annotation should be loaded. must match stream names in nova.
          start: optional start time_ms. use if only a specific chunk of a session should be retreived.
          end: optional end time_ms. use if only a specifc chunk of a session should be retreived.
          **kwargs: arguments that will be passed through to the dataset builder
        """
        self.dataset = dataset
        self.nova_data_dir = nova_data_dir
        self.sessions = sessions
        self.roles = roles
        self.schemes = schemes
        self.data_streams = data_streams
        self.annotator = annotator
        self.left_context_ms = ndu.parse_time_string_to_ms(left_context)
        self.right_context_ms = ndu.parse_time_string_to_ms(right_context)
        self.frame_size_ms = ndu.parse_time_string_to_ms(frame_size)
        self.stride_ms = ndu.parse_time_string_to_ms(stride) if stride else self.frame_size_ms
        self.start_ms = ndu.parse_time_string_to_ms(start) if start else 0
        self.end_ms = ndu.parse_time_string_to_ms(end) if end else float('inf')
        self.flatten_samples = flatten_samples
        self.clear_cache = clear_cache
        self.add_rest_class = add_rest_class
        self.lazy_loading = lazy_loading
        self.nova_db_handler = NovaDBHandler(db_config_path, db_config_dict)

        # Retrieving meta information from the database
        mongo_schemes = self.nova_db_handler.get_schemes(dataset=dataset, schemes=schemes)
        mongo_data = self.nova_db_handler.get_data_streams(dataset=dataset, data_streams=data_streams)
        self.label_info = self._populate_label_info_from_mongo_doc(mongo_schemes)
        self.data_info = self._populate_data_info_from_mongo_doc(mongo_data)

        # setting supervised keys
        if supervised_keys and self.flatten_samples:
            if supervised_keys[0] not in self.data_streams:
                # remove <role> of supervised keys
                _, data_stream = split_role_key(supervised_keys[0])
                if not data_stream in self.data_streams:
                    print('Warning: Cannot find supervised key \'{}\' in datastreams'.format(data_stream))
                    raise Warning('Unknown data_stream')
                else:
                    supervised_keys[0] = data_stream
            if supervised_keys[1] not in self.schemes:
                # remove <role> of supervised keys
                _, scheme = split_role_key(supervised_keys[1])
                if not scheme in schemes:
                    print('Warning: Cannot find supervised key \'{}\' in schemes'.format(scheme))
                    raise Warning('Unknown scheme')
                else:
                    supervised_keys[1] = scheme

        self.supervised_keys = tuple(supervised_keys) if supervised_keys else None

        super(HcaiNovaDynamic, self).__init__(**kwargs)

        if clear_cache:
            try:
                shutil.rmtree(self.data_dir)
            except OSError as e:
                print("Error: %s : %s" % (self.data_dir, e.strerror))

    def _info(self) -> tfds.core.DatasetInfo:
        """Returns the dataset metadata."""

        def map_label_id(lid):
            if self.flatten_samples and not lid == 'frame':
                return split_role_key(lid)[-1]
            return lid

        features_dict = {
                    # TODO: Remove frame when tfds implements option to disable shuffle
                    # Adding fake framenumber label for sorting
                    'frame': tf.string,
                    **{map_label_id(k): v.get_tf_info()[1] for k,v in self.label_info.items()},
                    **{map_label_id(k): v.get_tf_info()[1] for k, v in self.data_info.items()}
                }

        return tfds.core.DatasetInfo(
            builder=self,
            description=_DESCRIPTION,
            features=tfds.features.FeaturesDict(features_dict),
            supervised_keys= self.supervised_keys,
            homepage='https://github.com/hcmlab/nova',
            citation=_CITATION,
            # TODO: This option is currently disabled because it raises an error with tfds 4.3.0
            # Code should be updated once a newer version is released
            #disable_shuffling=True
        )

    def _populate_label_info_from_mongo_doc(self, mongo_schemes):
        """
        Setting self.label_info
        Args:
          mongo_schemes:

        Returns:

        """
        label_info = {}

        # List of all combinations from roles and schemes that occur in the retrieved data.
        for scheme in mongo_schemes:
            for role in self.roles:
                label_id = merge_role_key(role=role, key=scheme['name'])
                scheme_type = nt.string_to_enum(nt.AnnoTypes, scheme['type'])
                scheme_name = scheme['name']
                scheme_valid = scheme['isValid']

                if scheme_type == nt.AnnoTypes.DISCRETE:
                    labels = scheme['labels']
                    label_info[label_id] = nau.DiscreteAnnotation(role=role, add_rest_class=True, scheme=scheme_name, is_valid=scheme_valid, labels=labels)

                elif scheme_type == nt.AnnoTypes.DISCRETE_POLYGON:
                    labels = scheme['labels']
                    sr = scheme['sr']
                    label_info[label_id] = nau.DiscretePolygonAnnotation(role=role, scheme=scheme_name,
                                                                         is_valid=scheme_valid, labels=labels, sr=sr)

                else:
                    raise ValueError('Invalid label type {}'.format(scheme['type']))

        return label_info

    def _populate_data_info_from_mongo_doc(self, mongo_data_streams):
        """
        Setting self.data
        Args:
          mongo_schemes:

        Returns:

        """
        data_info = {}

        for data_stream in mongo_data_streams:
            for role in self.roles:
                sample_stream_name = role + '.' + data_stream['name'] + '.' + data_stream['fileExt']
                sample_data_path = os.path.join(self.nova_data_dir, self.dataset, self.sessions[0], sample_stream_name)
                dtype = nt.string_to_enum(nt.DataTypes, data_stream['type'])

                if dtype == nt.DataTypes.VIDEO:
                    data = VideoData(role=role, name=data_stream['name'], file_ext=data_stream['fileExt'], sr=data_stream['sr'], is_valid=data_stream['isValid'], sample_data_path=sample_data_path)
                elif dtype == nt.DataTypes.AUDIO:
                    raise NotImplementedError('Audio files are not yet supported')
                elif dtype == nt.DataTypes.FEATURE:
                    data = StreamData(role=role, name=data_stream['name'], sr=data_stream['sr'], data_type=dtype, is_valid=data_stream['isValid'], sample_data_path=sample_data_path, lazy_loading=self.lazy_loading)
                else:
                    raise ValueError('Invalid data type {}'.format(data_stream['type']))

                data_id = merge_role_key(role=role, key=data_stream['name'])
                data_info[data_id] = data

        return data_info

    def _split_generators(self, dl_manager: tfds.download.DownloadManager):
        """Returns SplitGenerators."""
        return {'dynamic_split': self._generate_examples()}

    # def _get_data_for_frame(self, file_reader, feature_type, sr, start, end):
    #     start_frame = ndu.milli_seconds_to_frame(sr, start)
    #     end_frame = ndu.milli_seconds_to_frame(sr, end)
    #     if feature_type == nt.DataTypes.VIDEO:
    #         return ndu.chunk_vid(file_reader, start_frame, end_frame)
    #     elif feature_type == nt.DataTypes.AUDIO:
    #         raise NotImplementedError('Data chunking for audio is not yet implemented')
    #     elif feature_type == nt.DataTypes.FEATURE:
    #         return ndu.chunk_stream(file_reader, start_frame, end_frame)

    def _load_annotation_for_session(self, session, time_to_ms=False):
        for label_id, anno in self.label_info.items():
            mongo_anno = self.nova_db_handler.get_annos(self.dataset, anno.scheme, session, self.annotator, anno.role)
            anno.set_annotation_from_mongo_doc(mongo_anno, time_to_ms=time_to_ms)

    def _open_data_reader_for_session(self, session):
        for data_id, data in self.data_info.items():
            data_path = os.path.join(self.nova_data_dir, self.dataset, session, data_id + '.' + data.file_ext)
            data.open_file_reader(data_path)

    def _generate_examples(self):
            """Yields examples."""

            # Needed to sort the samples later and assure that the order is the same as in nova. Necessary for CML till the tfds can be returned in order.
            sample_counter = 1

            for session in self.sessions:

                # Gather all data we need for this session
                self._load_annotation_for_session(session, time_to_ms=True)
                self._open_data_reader_for_session(session)

                session_info = self.nova_db_handler.get_session_info(self.dataset, session)[0]
                dur = int(min(*[v.dur for k,v in self.data_info.items()], session_info['duration'] ))

                if not dur:
                    raise ValueError('Session {} has no duration.'.format(session))

                dur_ms = dur * 1000

                # Starting position of the first frame in seconds
                c_pos_ms = self.left_context_ms + self.start_ms

                # Generate samples for this session
                while c_pos_ms + self.stride_ms + self.right_context_ms <= min(self.end_ms, dur_ms):
                    frame_start_ms = c_pos_ms - self.left_context_ms
                    frame_end_ms = c_pos_ms + self.frame_size_ms + self.right_context_ms
                    key = session + '_' + str(frame_start_ms / 1000) + '_' + str(frame_end_ms / 1000)

                    labels_for_frame = [{k: v.get_label_for_frame(frame_start_ms, frame_end_ms)} for k, v in self.label_info.items()]
                    data_for_frame = [{k: v.get_sample(frame_start_ms, frame_end_ms)} for k, v in self.data_info.items()]

                    sample_dict = {}

                    for l in labels_for_frame:
                        sample_dict.update(l)

                    for d in data_for_frame:
                        sample_dict.update(d)

                    if self.flatten_samples:

                        # grouping labels and data according to roles
                        for role in self.roles:

                            # filter dictionary to contain values for role
                            sample_dict_for_role = dict(filter(lambda elem: role in elem[0], sample_dict.items()))

                            # remove role from dictionary keys
                            sample_dict_for_role = dict(map(lambda elem: (elem[0].replace(role + '.', ''), elem[1]), sample_dict_for_role.items()))

                            sample_dict_for_role['frame'] = str(sample_counter) + '_' + key + '_' + role
                            yield key + '_' + role, sample_dict_for_role
                            sample_counter += 1
                        c_pos_ms += self.stride_ms

                    else:
                        sample_dict['frame'] = str(sample_counter) + '_' + key
                        yield key, sample_dict
                        c_pos_ms += self.stride_ms
                        sample_counter += 1

                # closing file readers for this session
                for k,v in self.data_info.items():
                    v.close_file_reader()



import tensorflow as tf
import tensorflow_datasets as tfds
import hcai_datasets.hcai_nova_dynamic.utils.nova_types as nt
import numpy as np

from abc import ABC, abstractmethod
from hcai_datasets.hcai_nova_dynamic.utils.nova_utils import merge_role_key, split_role_key

class Annotation(ABC):

    def __init__(self, role: str = '', scheme: str = '', is_valid: bool = True):
        self.role = role
        self.scheme = scheme
        self.is_valid = is_valid

        # Gets set when load_annotation is called
        self.data = None

    @abstractmethod
    def get_tf_info(self):
        """
        Returns the labels for this annotation to create the DatasetInfo for tensorflow
        """
        raise NotImplementedError

    @abstractmethod
    def set_annotation_from_mongo_doc(self, session, time_to_ms=False):
        """
        Returns the labels for this annotation to create the DatasetInfo for tensorflow
        """
        raise NotImplementedError

    @abstractmethod
    def get_label_for_frame(self, start, end):
        """
        Returns the label for this frame
        """
        raise NotImplementedError


class DiscreteAnnotation(Annotation):

    REST = 'REST'

    def __init__(self, labels= {}, add_rest_class=False, **kwargs):
        super().__init__(**kwargs)
        self.type = nt.AnnoTypes.DISCRETE
        self.labels = { x['id'] : x['name'] if x['isValid'] else '' for x in sorted(labels, key=lambda k: k['id']) }
        self.add_rest_class = add_rest_class
        if self.add_rest_class:
            self.labels[max(self.labels.keys()) + 1] = DiscreteAnnotation.REST

    def get_tf_info(self):
        return (merge_role_key(self.role, self.scheme), tfds.features.ClassLabel(names=list(self.labels.values())))

    def set_annotation_from_mongo_doc(self, mongo_doc, time_to_ms=False):
        self.data = mongo_doc
        if time_to_ms:
            for d in self.data:
                d['from'] = int(d['from'] * 1000)
                d['to'] = int(d['to'] * 1000)

    def get_label_for_frame(self, start, end):


        # If we don't have any data we return the garbage label
        if self.data == -1:
            return -1

        else:
            # Finding all annos that overlap with the frame
            def is_overlapping(af, at, ff, ft):

                # anno is larger than frame
                altf = af <= ff and at >= ft

                # anno overlaps frame start
                aofs = at >= ff and at <= ft

                # anno overlaps frame end
                aofe = af >= ff and af <= ft

                return altf or aofs or aofe

            annos_for_sample = list(filter(lambda x: is_overlapping(x['from'], x['to'], start, end), self.data))

            # No label matches
            if not annos_for_sample:
                if self.add_rest_class:
                    return len(self.labels.values()) -1
                else:
                    return -1

            majority_sample_idx = np.argmax(
                list(map(lambda x: min(end, x['to']) - max(start, x['from']), annos_for_sample)))

            return annos_for_sample[majority_sample_idx]['id']


class ContinousAnnotation(Annotation):

    def __init__(self, sr=0, min=0, max= 0, **kwargs):
        super().__init__(**kwargs)
        self.type = nt.AnnoTypes.CONTINOUS
        self.sr = sr
        self.min = min
        self.max = max


class FreeAnnotation(Annotation):

    def __init__(self, **kwargs):
        self.type = nt.AnnoTypes.FREE
        super().__init__(**kwargs)


class DiscretePolygonAnnotation(Annotation):

    def __init__(self, labels={}, sr=0, **kwargs):
        super().__init__(**kwargs)
        self.type = nt.AnnoTypes.DISCRETE_POLYGON
        self.labels = {str(x['id']): x['name'] if x['isValid'] else '' for x in sorted(labels, key=lambda k: k['id'])}
        self.sr = sr
        self.dummy_label = np.full( (10, 2), -1, dtype=np.float64)

    def get_tf_info(self):
        """
        Example:
           <role>.<scheme>, {
                'label_id_1' : [ (x1,y1), (x2,y2), ...  (xn, yn)],
                'label_id_2' : ...
                ...
,           }
        """
        tf_features = tfds.features.FeaturesDict({
            str(l): tfds.features.Sequence( tfds.features.Tensor(shape=(2, ), dtype=tf.float64)) for l in self.labels.keys()
        })
        return (merge_role_key(self.role, self.scheme), tf_features)

    def set_annotation_from_mongo_doc(self, mongo_doc, time_to_ms=False):
        self.data = mongo_doc

    def get_label_for_frame(self, start, end):
        # Use the start of the frame to determine the label
        frame_nr = int(self.sr * start)

        # Prefill array with dummy -1 labels
        label = {
            str(l): self.dummy_label for l in self.labels.keys()
            }

        # If we have any label data fill the label
        if not self.data == -1 and len(self.data) > frame_nr:
            for l in self.data[frame_nr]['polygons']:
                if str(l['label']) in label.keys():
                    label[str(l['label'])] = np.array( [(p['x'], p['y']) for p in l['points']] )
                else:
                    print('Warning! Label ID {} found in annotation but not in scheme.'.format(l['label']))
        return label










import xml.etree.ElementTree as ET
import numpy as np
from hcai_datasets.hcai_nova_dynamic.utils.ssi_data_types import NPDataTypes, TFDataTypes, FileTypes, string_to_enum

class Stream:
    #def __init__(self, ftype="UNDEF" , sr=0.0, dim=0, byte=4, type="UNDEF", delim="", chunks = []):
    def __init__(self, path=None):
        self.ftype = string_to_enum(FileTypes, "UNDEF")
        self.sr = 0
        self.dim = 0
        self.byte = 4
        self.type = "UNDEF"
        self.tftype = "UNDEF"
        self.delim = ""
        self.chunks = []
        self.data = None

        if path:
            self.load(path)

    def load_header(self, path):
        tree = ET.parse(path)
        root = tree.getroot()
        chunks = 0

        for child in root:
            if child.tag ==  'info':
                for key,val in child.attrib.items():
                    if key == 'ftype':
                        self.ftype = string_to_enum(FileTypes, val)
                    elif key == 'sr':
                        self.sr = float(val)
                    elif key == 'dim':
                        self.dim = int(val)
                    elif key == 'byte':
                        self.byte = int(val)
                    elif key == 'type':
                        self.tftype = string_to_enum(TFDataTypes, val).value
                        self.type = string_to_enum(NPDataTypes, val).value
                    elif key == 'delim':
                        self.delim = val
            elif child.tag == 'chunk':
                f, t, b, n = 0, 0, 0, 0
                for key,val in child.attrib.items():
                    if key == 'from':
                        f = float(val)
                    elif key == 'to':
                        t = float(val)
                    elif key == 'num':
                        n = int(val)
                    elif key == 'byte':
                        b = int(val)
                chunks += 1
                self.chunks.append( [f, t, b, n] )
        return self

    def load_data(self, path):
        if self.ftype == FileTypes.ASCII:
            self.data = np.loadtxt(path, dtype=self.type, delimiter=self.delim)
        elif self.ftype == FileTypes.BINARY:
            num = np.sum(self.chunks, axis=0, dtype=np.int32)[3]
            self.data = np.fromfile(path, dtype=self.type).reshape(num, self.dim)
        else:
            raise ValueError('FileType {} not supported'.format(self))
        return self.data

    def load(self, path):
        self.load_header(path)
        self.load_data(path + '~')
        return self

if __name__ == "__main__":
    stream_ascii = Stream().load("Testfiles/ascii.stream")
    stream_bin = Stream().load("Testfiles/binary.stream")
    


"""hcai_affectnet dataset."""

import tensorflow_datasets as tfds
import tensorflow as tf
import pandas as pd
import json
import numpy as np
from pathlib import Path
from tensorflow_datasets.core.splits import Split
from hcai_dataset_utils.statistics import Statistics


_DESCRIPTION = """
Affectnet is a dataset that has been crawled from the internet and annotated with respect to affective classes as well as valence and arousal.
The annotations also include automatically extracted facial landmarks.
Overall the dataset consists of roughly 1Million images. Half of the images are manually annotated where the other half has been annotated automatically.
Since there is currently no official test set available the validation set is used for testing and a split of the training set is used for validation.
The number of images in test is only 5499 since one corrupt image has been deleted
"""

_CITATION = """
@article{mollahosseini2017affectnet,
  title={Affectnet: A database for facial expression, valence, and arousal computing in the wild},
  author={Mollahosseini, Ali and Hasani, Behzad and Mahoor, Mohammad H},
  journal={IEEE Transactions on Affective Computing},
  volume={10},
  number={1},
  pages={18--31},
  year={2017},
  publisher={IEEE}
}
"""

VERSION = tfds.core.Version("1.0.0")
RELEASE_NOTES = {
    "1.0.0": "Initial release.",
}


class HcaiAffectnetConfig(tfds.core.BuilderConfig):
    """BuilderConfig for HcaiAffectnetConfig."""

    def __init__(
        self, *, include_auto=False, ignore_duplicate=True, ignore_lists=None, **kwargs
    ):
        """BuilderConfig for HcaiAffectnetConfig.
        Args:
          include_auto: bool. Flag to determine whether the automatically annotated files should be included in the dataset.
          include_auto: bool. Flag to determine whether the duplicated files in the dataset should be included.
          **kwargs: keyword arguments forwarded to super.
        """
        super(HcaiAffectnetConfig, self).__init__(version=VERSION, **kwargs)
        if ignore_lists is None:
            ignore_lists = []

        if ignore_duplicate:
            ignore_lists.append("affect_net_ignore_list_duplicates.json")
        self.include_auto = include_auto
        self.ignore_lists = ignore_lists


class HcaiAffectnet(tfds.core.GeneratorBasedBuilder, Statistics):
    """DatasetBuilder for hcai_affectnet dataset."""

    BUILDER_CONFIGS = [
        HcaiAffectnetConfig(name="default", include_auto=False, ignore_duplicate=True),
        HcaiAffectnetConfig(name="inc_auto", include_auto=True, ignore_duplicate=True),
    ]

    IMAGE_FOLDER_COL = "image_folder"

    def __init__(self, *, dataset_dir=None, **kwargs):
        self.dataset_dir = dataset_dir
        self.LABELS = [
            "neutral",
            "happy",
            "sad",
            "suprise",
            "fear",
            "disgust",
            "anger",
            "contempt",
            "none",
            "uncertain",
            "non-face",
        ]
        super(HcaiAffectnet, self).__init__(**kwargs)

    def _info(self) -> tfds.core.DatasetInfo:
        """Returns the dataset metadata."""
        return tfds.core.DatasetInfo(
            builder=self,
            description=_DESCRIPTION,
            metadata=tfds.core.MetadataDict({}),
            features=tfds.features.FeaturesDict(
                {
                    # These are the features of your dataset like images, labels ...
                    "image": tfds.features.Image(shape=(None, None, 3)),
                    "expression": tfds.features.ClassLabel(names=self.LABELS),
                    "arousal": tf.float32,
                    "valence": tf.float32,
                    "facial_landmarks": tfds.features.Tensor(
                        shape=(68, 2), dtype=tf.float32
                    ),
                    "rel_file_path": tf.string
                    #'face_bbox': tfds.features.BBox(),
                }
            ),
            supervised_keys=(
                "image",
                "expression",
            ),  # Set to `None` to disable
            homepage="https://dataset-homepage/",
            citation=_CITATION,
        )

    def _populate_meta_data(self, data: pd.DataFrame):
        """
        Updating the metadata dictionary
        Args:
            data: the dataframe that will be used for calulating the metadata
        """
        for split_name, df in data:
            # Filter only necessary columns
            df_filtered = df.filter(["expression", "arousal", "valence"])

            # Replacing -2 values with nan for the statistics
            df_filtered = df_filtered.applymap(lambda x: np.nan if x == -2 else x)

            # Converting integer to type object to enable correct description
            convert_dict = {"expression": "object"}
            df_filtered = df_filtered.astype(convert_dict)
            self._populate_stats(df_filtered, split_name)

    def _split_generators(self, dl_manager: tfds.download.DownloadManager):
        """Returns SplitGenerators."""
        print("Loading Labels...")

        train_csv_path = (
            Path(self.dataset_dir) / "Manually_Annotated_file_lists" / "training.csv"
        )
        test_csv_path = (
            Path(self.dataset_dir) / "Manually_Annotated_file_lists" / "validation.csv"
        )

        train_csv_path_auto = (
            Path(self.dataset_dir)
            / "Automatically_Annotated_file_lists"
            / "automatically_annotated.csv"
        )

        train_df = pd.read_csv(train_csv_path, index_col=0)
        test_df = pd.read_csv(test_csv_path, index_col=0)
        train_df[self.IMAGE_FOLDER_COL] = ["Manually_Annotated_Images"] * len(train_df)
        test_df[self.IMAGE_FOLDER_COL] = ["Manually_Annotated_Images"] * len(test_df)

        # append automatic labels if not ignored:
        if self.builder_config.include_auto:
            train_df_auto = pd.read_csv(train_csv_path_auto, index_col=0)
            train_df_auto[self.IMAGE_FOLDER_COL] = ["Automatically_Annotated_Images"]
            train_df = pd.concat([train_df, train_df_auto])

        len_train = len(train_df)
        len_test = len(test_df)
        print(
            "...loaded {} images for train\n...loaded {} images for test".format(
                len_train, len_test
            )
        )

        # removing labels that are specified in the ignore-lists
        print("Apply filtering...")

        filter_list_path = Path(__file__).parent / "Ignore_Lists"
        for filter_list in self._builder_config.ignore_lists:
            with open(filter_list_path / filter_list) as json_file:
                filter = json.load(json_file)
                train_df.drop(filter, errors="ignore", inplace=True)
                test_df.drop(filter, errors="ignore", inplace=True)
        print(
            "...dropped {} images from train\n... dropped {} images from test".format(
                len_train - len(train_df), len_test - len(test_df)
            )
        )

        # split train in validation an train
        """print("Splitting validation set")
        val_df = train_df.sample(frac=0.2, random_state=1337)
        train_df = train_df.drop(val_df.index)
        """
        print(
            "Final set sizes: \nTrain {}\nTest {}".format(len(train_df), len(test_df))
        )

        data = [(Split.TRAIN, train_df), (Split.TEST, test_df)]

        self._populate_meta_data(data)

        return {
            split_name: self._generate_examples(split_data)
            for split_name, split_data in data
        }

    def _generate_examples(self, label_df):
        for index, row in label_df.iterrows():
            yield index, {
                "image": Path(self.dataset_dir) / row[self.IMAGE_FOLDER_COL] / index,
                "expression": row["expression"],
                "arousal": row["arousal"],
                "valence": row["valence"],
                "facial_landmarks": np.fromstring(
                    row["facial_landmarks"], sep=";", dtype=np.float32
                ).reshape((68, 2)),
                "rel_file_path": row[self.IMAGE_FOLDER_COL] + "/" + index,
            }




def merge_role_key(key, role):
        return role + '.' + key

def split_role_key(label_key):
        split = label_key.split('.')
        role = split[0]
        key = '.'.join(split[1:])
        return role, key
